<?php
/**
 * Created by PhpStorm.
 * User: heyu
 * Date: 2019-05-16
 * Time: 14:34
 */

namespace kwy\mybank\request;

define('MYBANK_ENV_TEST','test');
define('MYBANK_ENV_PRODUCT','pro');
define('MYBANK_ENV_PRE_PRODUCT','pre_pro');

use kwy\mybank\exception\MybankException;
use kwy\mybank\exception\MyBankNetException;
use RobRichards\XMLSecLibs\XMLSecurityDSig;
use RobRichards\XMLSecLibs\XMLSecurityKey;

class MybankApi
{
    private $isvOrgId;

    private $appid;

    private $timeout=5;

    private $env;

    private $private_key;

    private $public_key;

    private $last_api_log;

    private $enableProxy=false;

    private $url=[
        'pro'=>'https://fcopen.mybank.cn/open/api/common/requestasync.htm',#生产环境
        'pre_pro'=>'https://fcopenpre.mybank.cn/open/api/common/requestasync.htm',#预发布环境 url
        'test'=>'https://fcsupergw.dl.alipaydev.com/open/api/common/requestasync.htm',#测试环境 url
    ];

    private $upload_url=[
        'pro'=>'https://fcsupergwlite.mybank.cn/ant/mybank/merchantprod/merchant/uploadphoto.htm',#生产环境
        'pre_pro'=>'https://fcsupergwlitepre.mybank.cn/ant/mybank/merchantprod/merchant/uploadphoto.htm',#预发布环境
        'test'=>'https://fcsupergwlite.dl.alipaydev.com/ant/mybank/merchantprod/merchant/uploadphoto.htm',#开发测试环境
    ];

    /**
     * @param mixed $isvOrgId
     */
    public function setIsvOrgId($isvOrgId)
    {
        $this->isvOrgId = $isvOrgId;
    }

    /**
     * @param mixed $appid
     */
    public function setAppid($appid)
    {
        $this->appid = $appid;
    }

    /**
     * @return mixed
     */
    public function getPrivateKey()
    {
        return $this->private_key;
    }

    /**
     * @param bool $enableProxy
     */
    public function setEnableProxy($enableProxy)
    {
        $this->enableProxy = $enableProxy;
    }



    /**
     * @param mixed $private_key
     */
    public function setPrivateKey($private_key)
    {
        $this->private_key = $private_key;
    }

    /**
     * @return mixed
     */
    public function getPublicKey()
    {
        return $this->public_key;
    }

    /**
     * @param mixed $public_key
     */
    public function setPublicKey($public_key)
    {
        $this->public_key = $public_key;
    }

    /**
     * @param mixed $env
     */
    public function setEnv($env)
    {
        $this->env = $env;
    }

    /**
     *
     * @param string $function
     * @param array $biz_params
     * @return mixed
     */
    protected function getBizResult($function,$biz_params){
        $req_log_str=json_encode($biz_params);
        foreach ($biz_params as &$v){
            if (is_array($v)){
                $v=base64_encode(json_encode($v));
            }
        }
        $xml_unsigned=$this->genRequestXml($function, $biz_params);
        $this->last_api_log='msgid:'.$this->getLastReqMsgId().'|function:'.$function.'|params:'.$req_log_str;
        $this->last_req_xml=$xml_signed=$this->signXml($xml_unsigned);

        try{
            $ret=$this->post($xml_signed);
            $this->last_api_log.='|resp:'.json_encode($ret);
        }catch (\Exception $e){
            $this->last_api_log.='|error_return:'.$e->getMessage();
            throw $e;
        }
        return $ret;

    }

    public function getLastApiLog(){
        return $this->last_api_log;
    }

    private function genRequestXml($function,$params){
        $head=$this->getHead($function);
        $params=array_merge($params,['IsvOrgId'=>$this->isvOrgId]);
        //$body=$this->getBody($params);
        $xml=$this->array2Xml(['document'=>['request'=>['head'=>$head,'body'=>$params]]]);
        return $xml;
    }

    public function getLastReqMsgId(){
        return $this->last_req_msg_id;
    }

    private function getRsaSign($params){
        ksort($params);
        $str='';
        foreach ($params as $k=>$v){
            if (!is_string($v)){
                continue;
            }
            $str.=$k.'='.$v.'&';
        }
        $str=rtrim($str,'&');
        $str=urlencode($str);

        $priKey=file_get_contents(__DIR__.'/cert/private_key.pem');
        $priRes=openssl_get_privatekey($priKey);
        $sign='';
        $ret=openssl_sign($str, $sign, $priRes,OPENSSL_ALGO_SHA256);
        openssl_free_key($priRes);
        $sign=base64_encode($sign);
        return $sign;
    }

    private function verifyRsaSign($params){
        $sign=$params['Signature'];
        unset($params['Signature']);
        ksort($params);
        $str='';
        foreach ($params as $k=>$v){
            if (is_string($v)){
                $str.=$k.'='.$v.'&';
            }
        }
        $str=trim($str,'&');
        //$str=urlencode(rtrim($str,'&'));

        $pubKey=file_get_contents(__DIR__.'/cert/public_key.pem');
        $Res=openssl_get_publickey($pubKey);
        $ret=(bool)openssl_verify($str, base64_decode($sign), $Res);
        openssl_free_key($Res);
        return $ret;
    }


    private function getHead($function){
        $this->last_req_msg_id=date('YmdHis').rand(10000,99999);
        return [
            'Appid'=>$this->appid,
            'Function'=>$function,
            'InputCharset'=>'UTF-8',
            'ReqTime'=>date('YmdHis'),
            'ReqMsgId'=>$this->last_req_msg_id,
            'SignType'=>'RSA',
            'Version'=>'1.0.0',
        ];
    }

    /**
     * 递归生成xml
     * @param unknown $array
     * @return unknown
     */
    public function array2Xml($array){
        $xml='';
        foreach ($array as $k=>$v){
            if (!is_array($v)){
                $xml.='<'.$k.'>'.$v.'</'.$k.'>'."\n";
            }else {
                $xml.='<'.$k.'>'."\n".$this->array2Xml($v).'</'.$k.'>'."\n";
            }
        }
        return $xml;
    }


    private function verifyResponseXml($xml){

        $xmlDocument = new \DOMDocument();
        $xmlDocument->loadXML($xml);
        $dsig=new XMLSecurityDSig();
        $dsig->locateSignature($xmlDocument);
        $dsig->canonicalizeSignedInfo();
        $dsig->locateKey();
        $dskey=new XMLSecurityKey(XMLSecurityKey::RSA_SHA1,['type'=>'public']);
        $dskey->loadKey($this->getPublicKey(),true);
        return $result = $dsig->verify($dskey);
    }

    protected function signXml($xml){

        $xmlDocument = new \DOMDocument();
        $xmlDocument->loadXML($xml);

        $dsig=new XMLSecurityDSig();
        $dsig->setCanonicalMethod(XMLSecurityDSig::C14N);
        $dsig->addReference($xmlDocument,
            XMLSecurityDSig::SHA1,
            ['http://www.w3.org/2000/09/xmldsig#enveloped-signature'],
            ['force_uri'=>true]
        );

        $dskey=new XMLSecurityKey(XMLSecurityKey::RSA_SHA1,['type'=>'private']);
        $dskey->loadKey($this->private_key,true);

        $dsig->sign($dskey);
        $dsig->appendSignature($xmlDocument->documentElement);

        //$xmlDocument->save('./signed.xml');
        //var_dump($xmlDocument->saveXML());
        return $xmlDocument->saveXML();

    }

    /**
     * 校验xml数字签名
     * @param unknown $xml
     */
    public function verifyXmlSign($xml){
        $xmlTool = new \XmlseclibsAdapter();
        if ($this->env=='test'){
            $mybank_pub_key=__DIR__ . '/cert/ALIBABA_BOGDA_PUBLIC_PRIVATE_CERT-pub-cer_2733.pem';
        }else {
            $mybank_pub_key=__DIR__ . '/cert/金融云-网商-开放平台-公钥-rsa2048-public_195493-生产.pem';
        }
        $xmlTool->setPublicKey(file_get_contents($mybank_pub_key));
        $xmlTool->addTransform(\XmlseclibsAdapter::ENVELOPED);
        $xmlTool->setKeyAlgorithm(\XMLSecurityKey::RSA_SHA256);
        $xmlDocument = new \DOMDocument();
        $xmlDocument->loadXML($xml);
        return $xmlTool->verify($xmlDocument);
    }

    private function getSgwError($code){
        $err_list=[
            'SGW01002'=>'系统未知异常',
            'SGW11009'=>'签名异常',
            'SGW12001'=>'接收到空报文',
            'SGW12014'=>'报文解析错误',
            'SGW12016'=>'验签错误',
            'SGW21008'=>'业务系统处理超时',
            'SGW13012'=>'同步/异步请求报文校验失败',
        ];
        if (isset($err_list[$code])){
            return 'SGW'.$err_list[$code];
        }else {
            return '网商未知错误:'.$code;
        }
    }

    public function post($data){
        $url=$this->url[$this->env];
        $api_req_time_b=microtime(true);
        $ch=curl_init();
        curl_setopt($ch, CURLOPT_URL, $url);//url
        curl_setopt($ch, CURLOPT_POST, 1);//post

        if (is_string($data)){
            curl_setopt($ch, CURLOPT_HTTPHEADER, array("Content-Type:text/xml; charset=utf-8"));
        }
        //curl_setopt($ch, CURLOPT_HTTPHEADER, ["Content-Type: multipart/form-data"]);

        curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);//返回结果curl
        curl_setopt($ch, CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_1);
        curl_setopt($ch, CURLOPT_IPRESOLVE, CURL_IPRESOLVE_V4 );//强制ipv4解析
        curl_setopt($ch, CURLOPT_POSTFIELDS, $data);
        curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, false);
        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);      //忽略校验证书
        curl_setopt($ch, CURLINFO_HEADER_OUT, true);
        curl_setopt($ch, CURLOPT_TIMEOUT, $this->timeout);

        #代理访问
        if ($this->enableProxy){
            curl_setopt($ch, CURLOPT_PROXY, "127.0.0.1"); //代理服务器地址
            curl_setopt($ch, CURLOPT_PROXYPORT,'8888'); //代理服务器端口
        }

        $ret = curl_exec($ch);

        if ($ret===false){
            throw new MyBankNetException('网络请求异常:'.curl_error($ch).',req_msgid:'.$this->getLastReqMsgId().','.$url,curl_errno($ch));
        }
        if (strpos($ret, 'SGW')===0){
            throw new MybankException($this->getSgwError($ret).',req_msgid:'.$this->getLastReqMsgId(),0,null,$ret);
        }
        $xmlDoc=new \DOMDocument();
        if ($xmlDoc->loadXML($ret)){
            $this->verifyResponseXml($ret);
            libxml_disable_entity_loader(true);
            $reponse_arr=json_decode(json_encode(simplexml_load_string($ret,null,LIBXML_NOCDATA)),true);
            return $reponse_arr['response']['body'];
        }else {
            if (strpos($ret, '<!DOCTYPE HTML PUBLIC')===0){
                throw new MybankException('网商系统故障，请稍后再试',0,null,$ret);
            }else {
                throw new MybankException($ret,0,null,$ret);
            }

        }

    }





    /**
     * 获取测试demo样例 请求xml数据内容
     * @return string
     */
    public static function getDemoFailTransferXml(){
        return file_get_contents(__DIR__.'/cert/fail_transfer_demo_request.xml');
    }


    /**
     * 设置接口请求超时时间
     * @param number $timeout
     */
    public function setTimeout($timeout){
        $this->timeout=$timeout;
    }

    /**
     * 生成业务外部交易号
     * @param string $prefix 交易号前缀
     */
    protected function genOutTradeNo($prefix=""){
        $this->last_out_trade_no=$prefix.date('YmdHis').rand(10000,99999);
        return $this->last_out_trade_no;
    }

    public function getLastOutTradeNo(){
        return $this->last_out_trade_no;
    }

    /**
     * 错误参数说明
     * ILLEGAL_PARAMETER("非法参数", "SGW810001"),
    MESSAGE_TRANSFORM_ERROR("报文转换错误", "SGW810002"),
    MESSAGE_TRANSFORM_NO_TEMPLATE("报文转换模板不存在", "SGW810003"),
    MESSAGE_TRANSFORM_EXCEPTION("报文转换异常", "SGW810004"),
    MESSAGE_TRANSFORM_NO_MSG("报文转换时入参无报文", "SGW810005"),
    MESSAGE_TRANSFORM_NO_CHANNEL_SYS_ID("报文转换时入参无渠道系统", "SGW810006"),
    MESSAGE_TRANSFORM_NO_DIRECTION_FLAG("报文转换时入参无方向标识", "SGW810007"),
    MESSAGE_TRANSFORM_NO_MSG_TYPE("报文转换时入参转无转换后的报文类型", "SGW810008"),
    MESSAGE_TRANSFORM_EMPTY("报文转换结果为空", "SGW810009"),
    MESSAGE_TRANSFORM_VTD_ERROR("报文转换VTD解析错误", "SGW810010"),
    MESSAGE_TRANSFORM_NOT_EXIST_TRANSFORMER("报文转换bean不存在", "SGW810011"),
    MESSAGE_TRANSFORM_INIT_EXCEPTION("报文转换bean不存在", "SGW810012"),
    MESSAGE_TRANSFORM_REPEAT_TEMPLATE("报文转换模板重复", "SGW810013"),
    MESSAGE_TRANSFORM_REPEAT_BEAN("报文转换模板bean重复", "SGW810014"),
    MESSAGE_TRANSFORM_NO_CACHE_KEY("报文转换模板没有生成缓存key", "SGW810015"),
    MESSAGE_TRANSFORM_NO_VERSION("报文转换时入参无版本号", "SGW810016"),
    MESSAGE_TRANSFORM_NO_FUNCTIONID("报文转换时入参无接口号", "SGW810017"),

    EMPTY_MESSAGE("空报文", "SGW12001"),
    BCM_FAIL("BCM处理失败", "SGW12002"),
    SYSTEM_ERROR("报文解析系统异常", "SGW12003"),
    SCRIPT_LOAD_ERROR("脚本加载错误", "SGW12004"),
    WRITE_SCRIPT_CONFIG_ERROR("写入脚本配置错误", "SGW12005"),
    SCRIPT_REMOVE_ERROR("脚本去除错误", "SGW12006"),
    ILLEGAL_XML("xml格式错误", "SGW12007"),
    ILLEGAL_JSON("json格式错误", "SGW12008"),
    ILLEGAL_PARAMETER("非法参数", "SGW12009"),
    TYPE_MISMATCH("解析器类型不匹配", "SGW12010"),
    NOT_EXIST_SCRIPT("解析脚本不存在", "SGW12011"),
    ILLEGAL_8583_BYTE("8583报文数据流格式错误", "SGW12012"),
    PARAM_NOT_FOUND("未定义报文参数", "SGW12013"),
    PARSE_ERROR("报文解析错误", "SGW12014"),
    DECRYPT_ERROR("解密错误", "SGW12015"),
    VERIFY_ERROR("验签错误", "SGW12016");

    NO_PARAM_NODE("param节点配置为空", "SGW13001"),
    NO_NAME_FIELD("field节点的name属性为空", "SGW13002"),
    ILLEGAL_NAME_FIELD("field节点的name属性配置不正确", "SGW13003"),
    NO_NAME_VALIDATOR("validator节点的name属性为空", "SGW13004"),
    ILLEGAL_XML("校验规则必须是合法的XML格式", "SGW13005"),
    ILLEGAL_IMPLEMENTATION("实现的不是Validator接口", "SGW13006"),
    BUILD_VALIDATOR_FAIL("组装校验器失败", "SGW13007"),
    NOT_REGISTED("没有注册的校验器", "SGW13008"),
    NONE_RULE("校验规则为空", "SGW13009"),
    ILLEGAL_PARAMETER("非法参数", "SGW13010"),
    VALIDATION_FAIL("校验失败", "SGW13011"),
    CALL_MESSAGE_VALIDATION_FAIL("同步/异步请求报文校验失败", "SGW13012"),
    RESULT_MESSAGE_VALIDATION_FAIL("同步响应报文/异步回调交易的请求报文校验失败", "SGW13013"),
    CALL_RESPONSE_VALIDATION_FAIL("异步请求交易的响应报文校验失败", "SGW13014"),
    RESULT_RESPONSE_VALIDATION_FAIL("异步回调交易的响应报文校验失败", "SGW13015"),
    UNIFY_COMMON_MESSAGE_NULL("统一支付报文2.0payload域为空", "SGW13016"),
    UNIFY_COMMON_MESSAGE_IS_NULL("统一支付报文2.0对象转换为空", "SGW13017"),
    WEB_SAFE_REFERER_FAIL("httprefer检查不通过", "SGW13018"),

    INNER_CODE_IS_NULL("主数据查询innerCode为空", "SGW13019");
     */
}